(require "iterate")

; case 1

(defconstant +escapes+
  (let ((entities '((#\< "&lt;")
                    (#\> "&gt;")
                    (#\& "&amp;")
                    (#\" "&quot;")
                    (#\€ "&euro;"))))
    (iter
      (with escapes = (make-array (1+ (iter
                                        (for e in entities)
                                        (maximizing (char-code (car e)))))
                                  :initial-element nil))
      (for e in entities)
      (setf (aref escapes (char-code (car e))) (cadr e))
      (finally (return escapes)))))

(defmacro aif (variable condition true false)
  `(let ((,variable ,condition))
     (if ,variable ,true ,false)))

(defun write-escaped (string &optional (stream *standard-output*))
  (iter
    (for char in-vector string)
    (let ((code (char-code char)))
      (aif escape (and (< code (length +escapes+))
                       (svref +escapes+ code))
        (write-string escape stream)
        (write-char   char   stream)))))

(defun print-html (html &optional (stream *standard-output*) indent)
  (labels ((format-value (stream value)
             (typecase value
               (string (write-escaped value stream))
               (symbol (format stream "~(~a~)" value))
               (t      (princ value stream))))
           (format-attribute (attribute value)
             (when value
               (format stream " ~(~a~)" attribute)
               (unless (eq value t)
                 (write-string "=\"" stream)
                 (if (consp value)
                   (iter
                     (for v = (pop value))
                     (when v (format-value stream v))
                     (while value)
                     (when v (write-char #\space stream)))
                   (format-value stream value))
                 (write-char #\" stream)))))
    (cond
      ((not html))
      ((not (listp html))
       (when indent (format stream "~v,0t" indent))
       (if (stringp html)
         (write-escaped html stream)
         (princ html stream))
       (when indent (terpri stream)))
      ((keywordp (car html))
       (let ((tag (pop html)))
         (case tag
           (:%raw
             (write-string (pop html) stream))
           (t
             (format stream "~v,0t<~(~a~)" (or indent 0) tag)
             (iter
               (while (keywordp (car html)))
               (format-attribute (pop html) (pop html)))
             (write-char #\> stream)
             (case (get tag 'kind)
               (void
                 (when html
                   (warn "HTML void tag <~(~a~)> has content: ~s" tag html)))
               (raw
                 (format stream "~{~@[~a~]~}</~(~a~)>" html tag))
               (escapable-raw
                 (iter
                   (for element in html)
                   (cond
                     ((stringp element)
                      (write-escaped element stream))
                     (element
                      (princ element stream))))
                 (format stream "</~(~a~)>" tag))
               (t
                 (if html
                   (iter
                     (with body-indent = (when indent (+ indent 2)))
                     (initially (when indent (terpri stream)))
                     (for element in html)
                     (print-html element stream body-indent)
                     (finally (format stream "~v,0t</~(~a~)>" (or indent 0) tag)))
                   (format stream "</~(~a~)>" tag))))
             (when indent (terpri stream))))))
      ((or (listp (car html)) (stringp (car html)))
       (iter
         (for x in html)
         (print-html x stream indent)))
      (t (error "Invalid HTML: ~s" html)))))

; case 2

(defclass backtrackable-input-stream (sb-gray:fundamental-input-stream)
  ((stream :initarg :stream)
   buffer))

(defmethod initialize-instance :after ((instance backtrackable-input-stream)
                                       &key &allow-other-keys)
  (setf (slot-value instance 'buffer)
        (make-array 0
                    :element-type (stream-element-type
                                    (slot-value instance 'stream))
                    :adjustable t
                    :fill-pointer t)))

(defmethod stream-element-type ((stream backtrackable-input-stream))
  (stream-element-type (slot-value stream 'stream)))

(defmethod close ((stream backtrackable-input-stream) &key abort)
  (close (slot-value stream 'stream) :abort abort))

(defmethod sb-gray:stream-file-position ((stream backtrackable-input-stream)
                                         &optional position-spec)
  (file-position (slot-value stream 'stream) position-spec))

(defmethod sb-gray:stream-clear-input ((stream backtrackable-input-stream))
  (setf (fill-pointer (slot-value stream 'buffer)) 0)
  (clear-input (slot-value stream 'stream)))

(defmethod sb-gray:stream-read-sequence ((stream backtrackable-input-stream) seq
                                         &optional (start 0) (end (length seq)))
  (with-slots (buffer stream) stream
    (iter
      (for i from start below end)
      (repeat (length buffer))
      (setf (aref seq i) (vector-pop buffer))
      (finally
        (return
          (if (< i end)
            (read-sequence seq (slot-value stream 'stream) :start i :end end)
            end))))))

(defclass backtrackable-character-input-stream (sb-gray:fundamental-character-input-stream
                                                backtrackable-input-stream)
  ())

(defmethod sb-gray:stream-peek-char ((stream backtrackable-character-input-stream))
  (with-slots (buffer stream) stream
    (if (plusp (length buffer))
      (aref buffer (1- (length buffer)))
      (peek-char nil stream nil :eof))))

(defmethod sb-gray:stream-read-char-no-hang ((stream backtrackable-character-input-stream))
  (with-slots (buffer stream) stream
    (if (plusp (length buffer))
      (vector-pop buffer)
      (read-char-no-hang stream nil :eof))))

(defmethod sb-gray:stream-read-char ((stream backtrackable-character-input-stream))
  (with-slots (buffer stream) stream
    (if (plusp (length buffer))
      (vector-pop buffer)
      (read-char stream nil :eof))))

(defmethod sb-gray:stream-read-line ((stream backtrackable-character-input-stream))
  (with-slots (buffer stream) stream
    (aif pos (position #\newline buffer :from-end t :test #'char=)
       (prog1 (nreverse (subseq buffer pos))
         (setf (fill-pointer buffer) pos))
       (if (plusp (length buffer))
         (concatenate 'string
                      (prog1 (reverse buffer)
                        (setf (fill-pointer buffer) 0))
                      (read-line stream nil :eof))
         (read-line stream nil :eof)))))

(defmethod sb-gray:stream-listen ((stream backtrackable-character-input-stream))
  (with-slots (buffer stream) stream
    (or (plusp (length buffer))
        (listen stream))))

(defmethod sb-gray:stream-unread-char ((stream backtrackable-character-input-stream) char)
  (vector-push-extend char (slot-value stream 'buffer)))

(defun make-backtrackable-stream (stream)
  (let ((type (stream-element-type stream)))
    (make-instance (if (subtypep type 'character)
                     'backtrackable-character-input-stream
                     (error "Backtrackable stream for ~s is not implemented" type))
                   :stream stream)))

(defgeneric unread-sequence (seq stream &key start end))

(defmethod unread-sequence (seq (stream backtrackable-input-stream)
                            &key (start 0) (end (length seq)))
  (let ((buffer (slot-value stream 'buffer)))
    (when (< (- (array-dimension buffer 0) (length buffer))
             (- end start))
      (adjust-array buffer (+ (array-dimension buffer 0) (- end start))))
    (iter
      (for x in-vector seq from (1- end) downto start)
      (vector-push x buffer))))

(defstruct template
  begin end map)

(defun str (&rest strings)
  (labels ((%str (x)
             (typecase x
               (string x)
               (list   (apply #'concatenate 'string (mapcar #'%str x)))
               (t      (format nil "~a" x)))))
    (%str strings)))

(defvar *templates*
  (list (make-template :begin "<%" :end "%>"
                       :map (lambda (forms) `(str ,.forms)))))

(define-compiler-macro str (&whole form &rest strings &environment env)
  (if (and (not (cdr strings))
           (constantp (car strings) env)
           (stringp (car strings)))
    (car strings)
    form))

(defun read-text (stream matches)
  "Read text from STREAM until a match is found. MATCHES are assumed to be sorted in decreasing length order. If there are several matches, the longest wins. The text up to this match is returned as the first value, and the match is returned as the second value. If the stream ends before any match, the second value is NIL."
  (let (match)
    (values
      (with-output-to-string (s)
        (let ((buffer (make-string (reduce #'max matches :key #'length))))
          (declare (dynamic-extent buffer))
          (iter read-loop
            (let ((c (peek-char nil stream nil)))
              (while c)
              (iter
                (with b = 0)
                (for m in matches)
                (when (char= c (char m 0))
                  (let ((l (length m)))
                    (when (and (or (>= b l)
                                   (= (setf b
                                            (read-sequence buffer stream
                                                           :start b
                                                           :end   l))
                                      l))
                               (string= buffer m :end1 l))
                      (setf match m)
                      (return-from read-loop))))
                (finally
                  (when (plusp b)
                    (unread-sequence buffer stream :end b))))
              (write-char (read-char stream) s)))))
      match)))

(defun skip-until (stream string)
  (iter
    (peek-char (char string 0) stream)
    (iter
      (for s in-vector string)
      (always (char= s (read-char stream)))
      (finally (return-from skip-until)))))

(defun skip-whitespace (stream)
  (iter
    (for c = (peek-char t stream))
    (case c
      (#\; (peek-char #\newline stream))
      (#\#
       (read-char stream)
       (if (char= (peek-char nil stream) #\|)
         (skip-until stream "|#")
         (progn
           (unread-char #\# stream)
           (return))))
      (t (return)))))

(defun read-fixed (stream string)
  (iter
    (for s in-vector string)
    (let ((c (read-char stream)))
      (unless (char= c s)
        (error "Expected '~c', got `~c' while reading `~a' from ~s"
               s c string stream))))
  string)

(defun read-forms (stream end-parenthesis)
  (iter
    (skip-whitespace stream)
    (while (iter
             (for p in-vector end-parenthesis with-index i)
             (for c = (read-char stream nil nil))
             (when (or (not c) (char/= c p))
               (unread-char c stream)
               (unread-sequence end-parenthesis stream :end i)
               (return t))))
    (collect (read stream))))

(defun read-template (stream &key (templates *templates*) extra-ends)
  (iter
    (with stream = (make-backtrackable-stream stream))
    (with ends = (sort (append extra-ends (mapcar #'template-begin templates))
                       #'>
                       :key #'length))
    (multiple-value-bind (text end) (read-text stream ends)
      (unless (string= text "") (collect text))
      (while end)
      (let ((template (find end templates :key #'template-begin :test #'eq)))
        (while template)
        (collect (funcall (template-map template)
                          (read-forms stream (template-end template))))))))

(defun load-template (pathname &key (templates *templates*))
  (with-open-file (f pathname)
    (read-template f :templates templates)))

(defun set-template-macro-character (begin-paren end-paren)
  (set-dispatch-macro-character #\# (char begin-paren 0)
    (lambda (stream char number)
      (declare (ignore number))
      (unread-char char stream)
      (read-fixed stream begin-paren)
      (cons 'str (read-template stream :extra-ends (list end-paren))))))

(set-template-macro-character "<%\"" "\"%>")

; document data

(defvar *metadata* "author")

(defun author (x) x)

; case 1 

(let ((doc `(:html
              (:head
                (:title "Hello Kitty")
                (:meta :name ,(author *metadata*)
                       :content "Not Sanrio Co., Ltd"))
              (:body
                (:h1 "Hello Kitty")
                (:p :class "official"
                    "She is not a cat. She is a human girl."))
              ,(iter
                 (repeat 3)
                 (collect `(:p :class "emphasis"
                               "Her name is Kitty White.")))
              (:p :class "citation-needed"
                  "We still don't know how she eats."))))
  (print-html doc *standard-output* 2))

; case 2

(let ((doc #<%"
              <html>
                <head>
                  <title>Hello Kitty</title>
                  <meta name="<% (author *metadata*) %>" content="Not Sanrio Co., Ltd"/>
                </head>
                <body>
                  <h1>Hello Kitty</h1>
                  <p class="official">She is not a cat. She is a human girl.</p>
                  <% (iter
                       (repeat 3)
                                ; an ordinary string would work here too
                       (collect #<%"<p class="emphasis">Her name is Kitty White.</p>"%>)) %>
                  <p class="citation-needed">We still don't know how she eats.</p>
                </body>
              </html>
            "%>))
  (princ doc))
